# See vtool dev/build_base_docker2.py
# See ~/code/vtool_ibeis/dev/build_base_docker2.py

"""
python ~/code/vtool_ibeis/dev/build_base_docker2.py --arch=x86_64 --parent_image_prefix=manylinux2014 --opencv
python ~/code/vtool_ibeis/dev/build_base_docker2.py --arch=i686 --parent_image_prefix=manylinux2014 --opencv
"""
