export VARNAME_CI_SECRET="EROTEMIC_CI_SECRET"
export VARNAME_TWINE_PASSWORD="EROTEMIC_PYPI_MASTER_TOKEN"
export VARNAME_TEST_TWINE_PASSWORD="EROTEMIC_TEST_PYPI_MASTER_TOKEN"
export VARNAME_TWINE_USERNAME="EROTEMIC_PYPI_MASTER_TOKEN_USERNAME"
export VARNAME_TEST_TWINE_USERNAME="EROTEMIC_TEST_PYPI_MASTER_TOKEN_USERNAME"
export GPG_IDENTIFIER="=Erotemic-CI <erotemic@gmail.com>"
